@@ -89,7 +89,7 @@ class ArmCompiler:
 
     # Override CCompiler.get_dependency_gen_args
     def get_dependency_gen_args(self, outtarget: str, outfile: str) -> T.List[str]:
-        return []
+        return ['--depend_target', outtarget, '--depend', outfile, '--depend_single_line']
 
     def get_pch_use_args(self, pch_dir: str, header: str) -> T.List[str]:
         # FIXME: Add required arguments
@@ -157,7 +157,7 @@ class ArmclangCompiler:
         self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage',
                              'b_ndebug', 'b_staticpic', 'b_colorout']
         # Assembly
-        self.can_compile_suffixes.update('s')
+        self.can_compile_suffixes.add('s')
 
     def get_pic_args(self) -> T.List[str]:
         # PIC support is not enabled by default for ARM,
