@@ -18,7 +18,7 @@ import shutil
 from . import ExtensionModule, ModuleReturnValue
 
 from .. import build, dependencies, mesonlib, mlog
-from ..interpreterbase import permittedKwargs, FeatureNew, stringArgs, InterpreterObject, ObjectHolder
+from ..interpreterbase import permittedKwargs, FeatureNew, stringArgs, InterpreterObject, ObjectHolder, noPosargs
 from ..interpreter import ConfigurationDataHolder, InterpreterException, SubprojectHolder
 
 
@@ -66,6 +66,7 @@ class CMakeSubprojectHolder(InterpreterObject, ObjectHolder):
                              'target': self.target,
                              'target_type': self.target_type,
                              'target_list': self.target_list,
+                             'found': self.found_method,
                              })
 
     def _args_to_info(self, args):
@@ -105,12 +106,18 @@ class CMakeSubprojectHolder(InterpreterObject, ObjectHolder):
         info = self._args_to_info(args)
         return info['func']
 
+    @noPosargs
     @permittedKwargs({})
     def target_list(self, args, kwargs):
-        if len(args) > 0:
-            raise InterpreterException('target_list does not take any parameters.')
         return self.held_object.cm_interpreter.target_list()
 
+    @noPosargs
+    @permittedKwargs({})
+    @FeatureNew('CMakeSubproject.found()', '0.53.2')
+    def found_method(self, args, kwargs):
+        return self.held_object is not None
+
+
 class CmakeModule(ExtensionModule):
     cmake_detected = False
     cmake_root = None
@@ -138,7 +145,7 @@ class CmakeModule(ExtensionModule):
         cmakebin = dependencies.ExternalProgram('cmake', silent=False)
         p, stdout, stderr = mesonlib.Popen_safe(cmakebin.get_command() + ['--system-information', '-G', 'Ninja'])[0:3]
         if p.returncode != 0:
-            mlog.log('error retrieving cmake informations: returnCode={0} stdout={1} stderr={2}'.format(p.returncode, stdout, stderr))
+            mlog.log('error retrieving cmake information: returnCode={0} stdout={1} stderr={2}'.format(p.returncode, stdout, stderr))
             return False
 
         match = re.search('\nCMAKE_ROOT \\"([^"]+)"\n', stdout.strip())
@@ -245,8 +252,7 @@ class CmakeModule(ExtensionModule):
         (ofile_path, ofile_fname) = os.path.split(os.path.join(state.subdir, '{}Config.cmake'.format(name)))
         ofile_abs = os.path.join(state.environment.build_dir, ofile_path, ofile_fname)
 
-        if 'install_dir' not in kwargs:
-            install_dir = os.path.join(state.environment.coredata.get_builtin_option('libdir'), 'cmake', name)
+        install_dir = kwargs.get('install_dir', os.path.join(state.environment.coredata.get_builtin_option('libdir'), 'cmake', name))
         if not isinstance(install_dir, str):
             raise mesonlib.MesonException('"install_dir" must be a string.')
 
